------------------------------common-----methods-----------------------------------------------
-- These methods are safe, by casting anything they got tostring() and by using our logging
-- system instead of stdout

local boundPrint = print
function print(...)
    local t = {}
    
    for i = 1, #arg do
	    table.insert(t, tostring(arg[i]))
    end

    local s = table.concat(t, "    ");
    boundPrint(s)
end

-- prints a table with an optional title
-- @param tbl - mandatory, the table to print
-- @param optional:title - defaults to nil
printTable = function(tbl, title)
	local output = {}
	
	for k,v in pairs(tbl) do
		table.insert(output, k .. " = " .. v)
	end

	if not title then 
		title = ""
	else
		title = title .. "\n"
	end

	print(title .. table.concat(output, ",\n"))
end

local boundError = error
error = function(msg, shouldThrow)

	local stackTrace = debug.traceback()

	if not shouldThrow then shouldThrow = false end

	boundError(msg .. "\n\n" .. stackTrace, shouldThrow)
end

-- prints message along with stacktrace
trace = function(msg)
	local stackTrace = debug.traceback()
	print(msg .. "\n\n" .. stackTrace)
end
------------------------------------------------------------------------------------------------


local ty, _types, global_tbl, idToTypesTable, engineParams = unpack(require 'engine.types')
local gm = require 'engine.manager'
local sprite = require 'engine.sprite'

----------------------------------------------------------------------------------------------- math methods implementation ---------------------------------------------------------------------
function round(num, idp)
  local mult = 10^(idp or 0)
  return math.floor(num * mult + 0.5) / mult
end

function round_noparams(num) 
	return math.floor(num+.5) 
end

-- Initialize the pseudo random number generator
-- See this URL: http://lua-users.org/wiki/MathLibraryTutorial
math.randomseed(os.time())
math.random()
math.random()
math.random()

function _random(arg1, arg2) 
	if arg1 and arg2 then
		return math.random(arg1, arg2)
	elseif arg1 then
		return math.random(0, arg1)
	else
		return math.random()
	end
end

function rad(degrees) 
	return degrees/180*math.pi
end

function _sqr(number) return number*number end 


function _tan(angle)
	return math.tan(rad(angle))
end

function _sin(angle)
	return math.sin(rad(angle))
end

function _cos(angle)
	return math.cos(rad(angle))
end


function _atan(tangent)
	return math.deg(math.atan(tangent))
end

function _asin(sine)
	return math.deg(math.asin(sine))
end

function _acos(cosine)
	return math.deg(math.acos(cosine))
end

function _avg(...)
	local sum = 0
	local numbers = {...}
	local length = #numbers
	for i=1, length do  sum=sum+numbers[i] end
	return sum/length
end
		
function lengthOfVector(x, y, x1, y1) 
	return math.sqrt((math.pow(x-x1,2) + math.pow(y-y1,2)))

end

function angleOfVector(x, y, x1, y1)
	return math.deg((math.atan2(x1-x, y1-y)))
end

-------------------------------------------------------------------- module's main ----------------------------------------------------------------



local global_mt = {}

function global_mt.__newindex(self, key, value)
	if ty[key] then
		error(key.." is a reserved word and cannot be assigned to")
	else
		
		local global_type = global_tbl[key]
		if global_type then
			if global_type == _types.globalIndicator then
				error(key.." is a reserved word and cannot be assigned to")
			elseif global_type == _types.globalVariable then
				local is_value_number = true
				if (value) then
					if (engineParams.DEBUG) then
						if (type(value) ~= "number") then
							error("Global variable value must be a number")
						elseif (value > engineParams.MAX_NUMBER or value < engineParams.MIN_NUMBER) then
							error("value exceeds allowed range [" .. engineParams.MIN_NUMBER .. ", " .. engineParams.MAX_NUMBER .. "]")
						end
					else
						if (type(value) ~= "number") then
							is_value_number = false
						elseif (value > engineParams.MAX_NUMBER) then
							value = engineParams.MAX_NUMBER
						elseif (value < engineParams.MIN_NUMBER) then
							value = engineParams.MIN_NUMBER
						end
					end
				elseif type(value) == "nil" then
					error("Cannot set global variable " .. tostring(key) .. " to a nil value!");
				end
				
				if (is_value_number == true) then
					gm.setGlobalVariable(key, value)
				end
--			elseif global_type == _types.builtinFunction then
--				error("cannot set variable!! global variable ".. key.. " is of a built-in function and cannot be overriden")
			else
				error("cannot set variable!! global variable ".. key.. " is of unexpected type")
			end
		else 
			rawset(_G, key, value)
		end
	end	
end

function global_mt.__index(self, key)
	local objectDesc = ty[key]
	if objectDesc then
		instances = gm.getInstances(objectDesc._id)
		return sprite:create(instances, key)
	else
		local global_type = global_tbl[key]
		if global_type then
			if global_type == _types.globalIndicator then
				return gm.getGlobalIndicator(key)
			elseif global_type == _types.globalVariable then
				return gm.getGlobalVariable(key)
			else
				error("cannot get variable!! global variable ".. key.. " is of unexpected type")
			end
		else

			-- Don't allow non existing vars in debug
			if (engineParams.DEBUG and not rawget(_G, key)) then

				local possiblyMatchingGlobalVar = nil
				for globalVarName, globalVarV in pairs(global_tbl) do
					if (string.lower(globalVarName) == string.lower(key)) then 
						possiblyMatchingGlobalVar = globalVarName
					end
				end
				if possiblyMatchingGlobalVar then
					error("The global variable '" .. key .. "' does not exist, did you mean '" .. possiblyMatchingGlobalVar .. "'?")
				else
					error("The global variable '" .. key .. "' does not exist")
				end 

				return nil
			else
				return rawget(_G, key)
			end
		end
		
		
	end
end

-------------------------------------------------------------------- global hook - game engine intergration --------------------------------------

rawset(_G,"tan", _tan)
rawset(_G,"sin", _sin)
rawset(_G,"cos", _cos)
rawset(_G,"atan", _atan)
rawset(_G,"asin", _asin)
rawset(_G,"acos", _acos)
rawset(_G,"abs", math.abs)
rawset(_G,"trunc", math.floor)
rawset(_G,"round", round_noparams)
rawset(_G,"sqr", _sqr)
rawset(_G,"sqrt", math.sqrt);
rawset(_G,"pow", math.pow)
rawset(_G,"rand", _random);

rawset(_G,"lengthOfVector", lengthOfVector);
rawset(_G,"angleOfVector", angleOfVector);
rawset(_G,"min", math.min);
rawset(_G,"max", math.max);
rawset(_G,"avg", _avg);


setmetatable(_G, global_mt)

return _G












