local networking = require 'engine.networking'
local utils = require 'engine.utils'

-- Local utilities

-- Stores bindings (eventid to function mapping)
local eventBindings = {}

--[[
	Handles all events. This routes calls to the function bound to events.
	This gets called by Java code (AndroidNetworkEngineBindings).
]]
function g_networkingEvents_globalEventHandler(eventId, args)
	local eventHandler = eventBindings[eventId]
	if eventHandler ~= nil then
		eventHandler(args)
	end
end

-- Registers g_networkingEvents_globalEventHandler to receive all network events.
local function registerGlobalEventHandler()
	--[[
		This may happen after NetworkingBindingsAbstract was reset (e.g. after
		changing rooms). Old event bindings should be forgotten.
	]]--
	if networking.getEventCallback() == nil then
		eventBindings = {}
		networking.registerEventCallback("g_networkingEvents_globalEventHandler")
	end
end

-- Public interface

networkingEvents = {}


--[[
	Sends a custom event to all players in the same room.

	All other players may react to this event if they previously called `bind`
	with the same eventId.

	The local player will not execute the event handler.

	@param eventId
			Identifier of the event.
	@param args
			Additional data.
]]
function networkingEvents.send(eventId, args)
	local keys, values = utils.toKeyAndValueArrays(args)
	networking.sendEventInternal(eventId, keys, values)
end

--[[ 
	Registers a function to be called when an event occurs.

	@param eventId 
			Identifier of the event.
	@param callback
			Function to be called when the event occurs. 
			the function receieves a single argument - a table whose keys and 
			values are strings, which contains event specific data.
			Note that this parameter is a function object not functio name.

	Example:
			networkingEvents.bind(123, function(args) 
				game.trace("got event 123: "..args)
			end)
]]
function networkingEvents.bind(eventId, callback)
	registerGlobalEventHandler()
	eventBindings[eventId] = callback
end

--[[
	Unregisters an event handler.

	@param eventId 
			Identifier of the event.
]]
function networkingEvents.unbind(eventId)
	eventBindings[eventId] = nil
end

--[[
	Unregsiter all events.
]]
function networkingEvents.unbindAll()
	eventBindings = {}
end

return networkingEvents