local ty, _types, globalsTable, idToTypesTable, engineParams = unpack(require 'engine.types')
local gm = require 'engine.manager'

-- gets the length of a Lua table. Note that the built-in # operator works for arrays only
local getTableLength = function (tbl)
	local count = 0

	for x,v in pairs(tbl) do
		count = count + 1
	end
	return count
end

--------------------- Text Frame proxy ----------------------------------------------------------------

local canvasItemProxy = {}
local canvasItem_mt = {}

function canvasItem_mt.__index(self, key)
	return gm.getCanvasItemParam(self.target_id, self.item_id, key)
end

function canvasItem_mt.__newindex(self, key, value) 
	if (self.lua_support) then
		gm.setCanvasItemParam(self.target_id, self.item_id, key, value)
	else
		error("Canvas item "..key.." is not supported for lua")
	end
end

function canvasItemProxy:create(sprite_id, canvasitem_id, is_lua_supported) 
	return setmetatable({ target_id = sprite_id, item_id=canvasitem_id, lua_support = is_lua_supported}, canvasItem_mt)
end

--------------------- Canvas Proxy ----------------------------------------------------------------------

local canvasProxy = {}
local canvasProxy_mt = {}

function canvasProxy_mt.__index(self, key)
	if (key == 'items') then return rawget(self,key) end 
		
	if self.items[key] then
		local itemId = self.items[key]._id
		local isLuaSupported = self.items[key]._luasupport
		local value = canvasItemProxy:create(self.target_id, itemId, isLuaSupported)
		return value
	else 
		error("canvas item: "..key.." does not exist")
	end
end

function canvasProxy_mt.__newindex(self, key, value) 
	error("canvas is immutable")
end

function canvasProxy:create(sprite_id, canvas_items) 
	return setmetatable({ target_id = sprite_id, items = canvas_items}, canvasProxy_mt)
end

--------------------- Custom Event Delegate ----------------------------------------------------------

local customEventDelegate = {}
local ce_mt = {}

function ce_mt.__call(self, args) 
	if args == nil then -- for custom events without args
        args = {}
	elseif type(args) ~= "table" then
		error("A custom event can only accept a table with it's params, given a " .. type(args), true)
    end

	-- validate the number of arguments and find the right custom event (overloading by the number of arguments is supported)
	local numArgs = getTableLength(args)
	local customEventDesc = self.custom_event_dispatch[numArgs]
	if not customEventDesc then 
		expectedNumArgs = 0

		-- There's only one key in the table so it's merely 1 iteration
		for k,v in pairs(self.custom_event_dispatch) do
			expectedNumArgs = k
		end
		error ("The custom event " .. tostring(self.name) .. 
			   " was given invalid number of arguments, got: "  ..tostring(numArgs) .. 
			   " expected: " .. tostring(expectedNumArgs) .. 
			   "\n\nTrace: " .. debug.traceback())
	end
	
	for key,value in pairs(args) do 
		if (type(value) == "number") then
			if (engineParams.DEBUG) then
				if (value > engineParams.MAX_NUMBER or value < engineParams.MIN_NUMBER) then
					error("value exceeds allowed range [" .. engineParams.MIN_NUMBER .. ", " .. engineParams.MAX_NUMBER .. "]")
				end
			else
				if (value > engineParams.MAX_NUMBER) then
					args[key] = engineParams.MAX_NUMBER
				end

				if (value < engineParams.MIN_NUMBER) then
					args[key] = engineParams.MIN_NUMBER
				end
			end
		end
	end
	
	gm.callCustomEvent({self.target_id}, customEventDesc._id, args)
end

function customEventDelegate:create(sprite_id, ce_dipatchTable, name) 
	return setmetatable({ target_id = sprite_id, custom_event_dispatch=ce_dipatchTable, name = name}, ce_mt)
end

		
--------------------- Sprite ----------------------------------------------------------

local SpriteRepo = {}
local Sprite = {}
local mt = {}



local properties = { 
			XSpeed = {gm.getSpriteSpeedX, gm.setSpriteSpeedX},
			YSpeed = {gm.getSpriteSpeedY, gm.setSpriteSpeedY},
			XPos = {gm.getSpritePosX, gm.setSpritePosX},
			YPos = {gm.getSpritePosY, gm.setSpritePosY},
			XAABB = {gm.getSpriteXAABB},
			YAABB = {gm.getSpriteYAABB},
			XScale = {gm.getSpriteScaleX, gm.setSpriteScaleX},
			YScale = {gm.getSpriteScaleY, gm.setSpriteScaleY},
			Width = {gm.getSpriteWidth},
			Height = {gm.getSpriteHeight}
		}

local setter_methods = {
		}

function mt.__newindex(self,key, value)
	rk = '___'..key
	local is_value_number = true

	if (value) then
		if (engineParams.DEBUG) then
			if (type(value) ~= "number") then
				error("numeric properties value must be a number")
			elseif (value > engineParams.MAX_NUMBER or value < engineParams.MIN_NUMBER) then
				error("value exceeds allowed range [" .. engineParams.MIN_NUMBER .. ", " .. engineParams.MAX_NUMBER .. "]")
			end
		else
			if (type(value) ~= "number") then
				is_value_number = false
			elseif (value > engineParams.MAX_NUMBER) then
				value = engineParams.MAX_NUMBER
			elseif (value < engineParams.MIN_NUMBER) then
				value = engineParams.MIN_NUMBER
			end
		end

	end

	if (is_value_number == true) then
		-- first get try to retrieve the built-in property return if available
		local exists = gm.getLocalIndicator(self.id, key)

		if exists then
			gm.setLocalIndicator(self.id, key, value)
		else
			-- second get try to retrieve setter for build-in property or setter method and invoke it if available
			local set_method =  (properties[key] and properties[key][2]) or setter_methods[key]
			if set_method then
				value = set_method(self.id, value)
				exists = true
			-- else if the key is a property of this sprite set it. Note that only numeric popery assingment are supported
			elseif ty[self.type].properties[key] then
				prop = ty[self.type].properties[key]
			
				if prop._t == _types.numericProperty then
					value = gm.setNumericProp(self.id, prop._id, value, prop._evId)
					exists = true
				else 
					error("Assignmnets to properties that are not of type 'numeric' is not supported: " .. key .. 
						"\n\nTrace: " .. debug.traceback())
				end
			end
		end

		if not exists then
			error('Cannot set member. Invalid indicator: '..key)
		end

		rawset(self, rk, value)
	end
end

function mt.__index(self, key)
	if (key == 'id') or (key=='type') then
		return rawget(self,key)
	end
	
	rk = '___'..key
	
	local value = rawget(self,rk)
	if value then
		return value
	end

	value = gm.getLocalIndicator(self.id, key)

	if not value then
		local get_method = properties[key] and properties[key][1]

		if get_method then
			value = get_method(self.id)
		end
	end
	
	if not value and ty[self.type].properties[key] then
		prop = ty[self.type].properties[key]
		if prop then
			if prop._t == _types.numericProperty then
				value = gm.getNumericProp(self.id, prop._id)
			
			elseif prop._t == _types.instanceProperty then
				local instances = gm.getInstanceProp(self.id, prop._id)
				if instances then
					value = SpriteGroup:create(instances, ty[self.type].properties[key]['_co'])
				end
			end
		end
		
	elseif ty[self.type].customEvents[key] then
		-- get the dispatch table and pass it as an argument to the delegate
		customEventDispatchTable = ty[self.type].customEvents[key]
		value = customEventDelegate:create(self.id, customEventDispatchTable, key)

	elseif key == "canvas" then
		value = canvasProxy:create(self.id, ty[self.type].textFrames)
	else
		if ty[key] then 
			value = Sprite:create(self.id, key) 
		end
	end
	
	if not value then
		error('Cannot get member. Invalid key name: '..key)
	end
	
	rawset(self, rk, value)
	return value		

end

function Sprite:create(sprite_id, type_id)
	local key = type_id..sprite_id
	
	if not SpriteRepo[key] then
		SpriteRepo[key] = setmetatable({ id = sprite_id, type=type_id}, mt)
	end
	
	return SpriteRepo[key]
end

------------------------------------------- Sprite Group -------------------------------------------------------------------------------

SpriteGroup = {}
SpriteGroup_mt = {}

function SpriteGroup:clearcache()
	SpriteRepo = {}
end

--[[
Instansiates a SpritesGroup which is an immutable array of Sprite instances.

@param [table array|number] sprites - engine sprite instances ids to wrap 
                                      the instansiated group around.
@param [int|string|table]type_id - GameObject type of the wrapped instances:
			                       [int] - id of the given type which corresponds to ty["GameObjectName"]._id
			                       [string] - the string which corresponds to "GameObjectName" in ty
			                       [table] - list of [int] (see above)
]]--
function SpriteGroup:create(sprites, type_id)

	if (type(sprites) == "number") then
		sprites = { sprites }
	end

	assert(#sprites >= 0, "sprite list number cannot be negative member")
	
	list = {}

	for i = 1, #sprites do

		tid = type_id

		if (type(type_id) == "number") then
			tid = idToTypesTable[type_id]
		elseif (type(type_id) == "table") then
			tid = idToTypesTable[type_id[i]] -- type id is actually the string representing the type
		end

		table.insert(list, Sprite:create(sprites[i], tid))

	end
	
	return setmetatable({_sprites = list } ,SpriteGroup_mt )
end

--[[
@param string key - if string changes properties of the sprite who's key is [key] to
                     the given value

Examples:

-- assume we have 3 sprites in the room which are of type mySprite:
mySprite.XPos = 5 -- set all sprites XPos to 5

mySprite[1] = 4 -- error! a sprite group is immutable

]]--
function SpriteGroup_mt.__newindex(self,key, value)
	local index = key
	if (type(index) ~= "number") then
		for i=1, #self._sprites do
			self._sprites[i][key] = value
		end
	else
		error("Sprite groups are immutable - cannot add values to the group")
	end
end

-- Returns the amount of sprites in the group
function SpriteGroup_mt:Length(self)
	return #self._sprites
end

function SpriteGroup_mt.__index(self, key)
	local index = key
	local sp = rawget(self,"_sprites")

	if key=='Length' then
		return SpriteGroup_mt:Length(self)
	
	elseif (type(index) ~= "number") then
		if #sp>0 then 
			return sp[1][key]
		else
			return nil
		end
	elseif (type(index) == "number") then
		if index >= 1 and index <= #sp then
			return sp[index]
		else
			error("Given sprite index is invalid: " .. tostring(index) ..
				  "\n\nTrace: " .. debug.traceback())
		end
	end

	return sp[key]
end


return SpriteGroup